from __future__ import annotations

from functools import partial
from typing import Callable, Sequence, Tuple

import flax.linen as nn
import jax
import jax.numpy as jnp

from ..ops import distance_similarity, non_max_suppression
from ..typing import Array, ArrayLike
from .common import DefaultUnpicklerMixin


def generate_predictions(module, outputs):
    """
    Produce a list of proposal locations based on predication map, remove redundency with non_max_suppression
    """
    locations = outputs["pred_locs"]
    scores = jax.nn.sigmoid(outputs["logits"])

    distance_threshold = module.nms_threshold
    output_size = module.max_output
    topk = module.pre_nms_topk
    score_threshold = module.min_score

    # sort and nms
    if topk <= 0 or topk > scores.size:
        topk = scores.size

    scores, indices = jax.lax.top_k(scores, topk)
    locations = locations[indices]

    threshold = 1 / distance_threshold / distance_threshold
    sel = non_max_suppression(
        scores,
        locations,
        output_size,
        threshold,
        score_threshold,
        return_selection=True,
    )

    idx_of_selected = jnp.argwhere(sel, size=output_size, fill_value=-1).squeeze(-1)

    scores = jnp.where(idx_of_selected >= 0, scores[idx_of_selected], -1.0)
    locations = jnp.where(
        idx_of_selected[:, None] >= 0, locations[idx_of_selected], -1.0
    )

    return dict(
        scores=scores,
        locations=locations,
    )


class LPN(nn.Module, DefaultUnpicklerMixin):
    """Location detection head

    Attributes:
        nms_threshold: non-max-supression threshold, if performing nms on detected locations.
        pre_nms_topk: max number of detections to be processed regardless of nms, ignored if negative
        max_output: number of detection outputs
    """

    # network hyperparams
    feature_scale: int = 4
    dtype: Any = None

    # detection hyperparams
    nms_threshold: float = 8.0
    pre_nms_topk: int = -1
    max_output: int = 512
    min_score: float = 0.2

    @nn.compact
    def _block(self, feature: ArrayLike) -> dict:
        x = feature
        height, width = x.shape[:-1]

        logits = nn.Conv(1, (1, 1), dtype=self.dtype)(x)
        regressions = nn.Conv(2, (1, 1), dtype=self.dtype)(x)

        ref_locs = (
            jnp.moveaxis(jnp.mgrid[:height, :width] + 0.5, 0, -1) * self.feature_scale
        )
        locs = ref_locs + regressions * self.feature_scale

        return dict(
            regressions=regressions.reshape(-1, 2),
            logits=logits.reshape(-1),
            ref_locs=ref_locs.reshape(-1, 2),
            pred_locs=locs.reshape(-1, 2),
        )

    def __call__(self, feature: ArrayLike, mask: ArrayLike | None = None) -> dict:
        network_outputs = self._block(feature)
        predictions = generate_predictions(self, network_outputs)

        if mask is not None:
            pred_locs = tuple(
                jnp.floor(predictions["locations"]).transpose().astype(int)
            )
            masked = mask[pred_locs]
            predictions["scores"] = jnp.where(masked, predictions["scores"], 0)
            predictions["locations"] = jnp.where(
                masked[:, None], predictions["locations"], -1
            )

        return dict(
            detector=network_outputs,
            predictions=predictions,
        )



from __future__ import annotations

from dataclasses import field
from functools import partial

import flax.linen as nn
import jax

from ..typing import Any, Array, ArrayLike
from ..utils import deep_update
from .backbone import Backbone
from .common import DefaultUnpicklerMixin
from .lpn import LPN
from .lpn3d import LPN3D
from .segmentor import Segmentor
from .segmentor3d import Segmentor3D

jnp = jax.numpy


class Lacss(nn.Module, DefaultUnpicklerMixin):
    """Main class for LACSS model

    Attributes:
        backbone: CNN backbone
        detector: Detection head to predict cell locations
        segmentor: Segmentation head
        detector_3d: Detection head for 3d input
        segmentor_3d: Segmentation head for 3d input
    """

    backbone: nn.Module = field(default_factory=Backbone)
    detector: nn.Module = field(default_factory=LPN)
    segmentor: nn.Module | None = field(default_factory=Segmentor)
    detector_3d: nn.Module | None = None
    segmentor_3d: nn.Module | None = None

    def get_features(self, image, ref_features, deterministic):
        return self.backbone(image, ref_features, deterministic=deterministic)

    def __call__(
        self,
        image: ArrayLike,
        *,
        image_mask: ArrayLike | None = None,
        video_refs: tuple | None = None,
    ) -> dict:
        """
        Args:
            image: [H, W, C] or [D, H, W, C]
        Returns:
            a dict of model outputs

        """
        # add a batch dim if needed
        image = jnp.asarray(image)

        x_det, x_seg = self.get_features(image, video_refs, deterministic=True)
        outputs = dict(
            det_features=x_det,
            seg_features=x_seg,
        )

        if image.ndim == 3:
            detector_outputs = self.detector(x_det, mask=image_mask)
            outputs = deep_update(outputs, detector_outputs)

            if self.segmentor is not None:
                segmentor_out = self.segmentor(
                    x_seg,
                    outputs["predictions"]["locations"],
                )
                outputs = deep_update(outputs, segmentor_out)

        else:
            if self.detector_3d is None:
                raise (ValueError("not a 3d detector"))

            detector_outputs = self.detector_3d(x_det, mask=image_mask)
            outputs = deep_update(outputs, detector_outputs)

            if self.segmentor_3d is not None:
                # segmentor_out = self.segmentor_3d(
                # x_seg, outputs["predictions"]["locations"],
                # )

                locs = outputs["predictions"]["locations"]
                locs = locs.reshape((locs.shape[0] // 16, 16) + locs.shape[1:])
                _, segmentor_out = jax.lax.scan(
                    lambda _, locs_: (None, self.segmentor_3d(x_seg, locs_)),
                    None,
                    locs,
                )
                segmentor_out = jax.tree_util.tree_map(
                    lambda data: data.reshape(-1, *data.shape[2:]),
                    segmentor_out,
                )

                outputs = deep_update(outputs, segmentor_out)

        return outputs

    def get_config(self):
        from dataclasses import asdict

        from ..utils import remove_dictkey

        cfg = asdict(self)

        remove_dictkey(cfg, "parent")

        return cfg

    @classmethod
    def from_config(cls, config):
        def _build_sub_module(module, k):
            if k in config and config[k] is not None:
                return module(**config[k])
            else:
                return None

        if isinstance(config, str):
            return cls.get_preconfigued(config)

        else:
            backbone = _build_sub_module(Backbone, "backbone")
            detector = _build_sub_module(LPN, "detector")

            assert backbone is not None and detector is not None, f"invalid config"

            obj = Lacss(
                backbone=backbone,
                detector=detector,
                segmentor=_build_sub_module(Segmentor, "segmentor"),
                detector_3d=_build_sub_module(LPN3D, "detector_3d"),
                segmentor_3d=_build_sub_module(Segmentor3D, "segmentor_3d"),
            )

            return obj

    @classmethod
    def get_default_model(cls, dtype=None):
        return cls.get_base_model(dtype=dtype)

    @classmethod
    def get_small_model(cls, dtype=None):
        return cls(
            backbone=Backbone(
                "tiny",
                fpn_dim=384,
                out_dim=256,
                dtype=dtype,
            ),
            detector=LPN(dtype=dtype),
            segmentor=Segmentor(
                patch_dim=32,
                sig_dim=512,
                pos_emb_shape=(16, 16, 4),
                dtype=dtype,
            ),
            detector_3d=LPN3D(dim=384, dtype=dtype),
            segmentor_3d=Segmentor3D(
                patch_dim=48,
                sig_dim=1024,
                pos_emb_shape=(8, 8, 8, 4),
                dtype=dtype,
            ),
        )

    @classmethod
    def get_base_model(cls, dtype=None):
        return cls(
            backbone=Backbone("base_v2", fpn_dim=512, out_dim=384, dtype=dtype),
            detector=LPN(dtype=dtype),
            segmentor=Segmentor(
                patch_dim=64,
                sig_dim=768,
                pos_emb_shape=(16, 16, 6),
                dtype=dtype,
            ),
            detector_3d=LPN3D(dim=512, dtype=dtype),
            segmentor_3d=Segmentor3D(
                patch_dim=64,
                sig_dim=1384,
                pos_emb_shape=(8, 8, 8, 6),
                dtype=dtype,
            ),
        )

    @classmethod
    def get_large_model(cls, dtype=None):
        return cls(
            backbone=Backbone("large_v2", fpn_dim=768, out_dim=512, dtype=dtype),
            detector=LPN(dtype=dtype),
            segmentor=Segmentor(
                patch_dim=96,
                sig_dim=1024,
                pos_emb_shape=(16, 16, 8),
                dtype=dtype,
            ),
            detector_3d=LPN3D(dim=768, dtype=dtype),
            segmentor_3d=Segmentor3D(
                patch_dim=96,
                sig_dim=2048,
                pos_emb_shape=(8, 8, 8, 8),
                dtype=dtype,
            ),
        )

    @classmethod
    def get_preconfigued(cls, config: str, dtype=None):
        if config == "default" or config == "base":
            return cls.get_default_model(dtype=dtype)
        elif config == "small":
            return cls.get_small_model(dtype=dtype)
        elif config == "large":
            return cls.get_large_model(dtype=dtype)
        else:
            raise ValueError(f"Unkown model config {config}")



from .auxiliary import *
from .common import *
from .instance import *


def lpn_loss(batch, prediction):
    return (
        prediction["losses"]["lpn_detection_loss"]
        + prediction["losses"]["lpn_localization_loss"]
    )



from __future__ import annotations

from functools import partial
from typing import Any

import flax.linen as nn
import jax
import optax
from xtrain.base_trainer import TrainIterator
from xtrain.strategy import LossLog, TrainState, VMapped
from xtrain.utils import Inputs, unpack_prediction_and_state, unpack_x_y_sample_weight

from ..utils import deep_update

jnp = jax.numpy


def lacss_mt_train_fn(module, image, teacher_prediction):
    n_train_instances = 16

    seg_locs = teacher_prediction["predictions"]["locations"][:n_train_instances]

    cutout_locs = (
        jax.random.uniform(module.make_rng("dropout"), shape=seg_locs.shape) * 50
        - 25
        + seg_locs
    )
    cutout_locs = jnp.where(seg_locs >= 0, cutout_locs.astype("int32"), 9999)
    image, _ = jax.lax.scan(
        lambda img, loc: (
            jax.lax.dynamic_update_slice(
                img,
                jnp.zeros([20, 20, 3], img.dtype),
                (loc[0] - 10, loc[1] - 10, 0),
            ),
            None,
        ),
        image,
        cutout_locs,
    )

    x_det, x_seg = module.get_features(image, None, deterministic=False)
    outputs = module.detector(x_det)

    pred_cls_prob = jax.nn.sigmoid(outputs["detector"]["logits"])
    gt_cls_prob = jax.nn.sigmoid(teacher_prediction["detector"]["logits"])
    mt_det_loss = pred_cls_prob * (1 - gt_cls_prob) + (1 - pred_cls_prob) * gt_cls_prob
    mt_det_loss = mt_det_loss * jnp.abs(gt_cls_prob - 0.5)
    mt_det_loss = mt_det_loss.sum() / jnp.abs(gt_cls_prob - 0.5).sum()

    pred_regr = outputs["detector"]["regressions"]
    gt_regr = teacher_prediction["detector"]["regressions"]
    mt_loc_loss = optax.l2_loss(pred_regr / 4, gt_regr / 4) * gt_cls_prob[:, None]
    mt_loc_loss = mt_loc_loss.sum() / gt_cls_prob[:, None].sum()

    segmentor_out = module.segmentor(x_seg, seg_locs)
    outputs = deep_update(outputs, segmentor_out)

    pred_segmentations = jax.nn.sigmoid(segmentor_out["predictions"]["segmentations"])
    gt_segmentations = jax.nn.sigmoid(
        teacher_prediction["predictions"]["segmentations"][:n_train_instances]
    )
    # mt_seg_loss = pred_segmentations * (1 - gt_segmentations) + (1 - pred_segmentations) * gt_segmentations
    mt_seg_loss = optax.l2_loss(pred_segmentations, gt_segmentations) * jnp.abs(
        gt_segmentations - 0.5
    )
    mt_seg_loss = mt_seg_loss.sum() / jnp.abs(gt_segmentations - 0.5).sum()

    outputs["losses"] = dict(
        mt_det_loss=mt_det_loss,
        mt_loc_loss=mt_loc_loss,
        mt_seg_loss=mt_seg_loss,
    )

    return outputs


def lacss_mt_loss(params, teacher_params, train_obj, inputs):
    inputs = Inputs.from_value(inputs)
    batch_size = jax.tree_util.tree_leaves(inputs)[0].shape[0]

    inputs = inputs.update(
        rngs={
            name: jax.random.split(jax.random.fold_in(rng, train_obj.step), batch_size)
            for name, rng in train_obj.rngs.items()
        },
        teacher_prediction=jax.vmap(
            Inputs.apply(
                train_obj.ctx.model.apply,
                dict(params=teacher_params),
            )
        )(inputs),
    )

    model_out = jax.vmap(
        Inputs.apply(
            nn.apply(lacss_mt_train_fn, train_obj.ctx.model),
            dict(params=params),
        )
    )(inputs)

    total_loss = sum(
        [
            model_out["losses"]["mt_det_loss"].sum(),
            model_out["losses"]["mt_seg_loss"].sum(),
            model_out["losses"]["mt_loc_loss"].sum(),
        ]
    )

    return total_loss, model_out


class MeanTeacher(VMapped):
    alpha: float = 0.05
    var_key: str = "teacher"

    @classmethod
    def train_step(
        cls,
        train_obj: TrainIterator,
        batch: Any,
    ) -> tuple[Any, Any]:
        inputs, label, _ = unpack_x_y_sample_weight(batch)

        if not cls.var_key in train_obj.variables:
            train_obj.variables[cls.var_key] = dict(
                params=train_obj.parameters,
                mt_loss_log=[
                    LossLog("losses/mt_det_loss"),
                    LossLog("losses/mt_seg_loss"),
                    LossLog("losses/mt_loc_loss"),
                ],
            )

        teacher_vars = train_obj.variables.pop(cls.var_key)

        if label is not None:
            prediction, train_obj = VMapped.train_step(train_obj, batch)

        else:
            grads, prediction = jax.grad(lacss_mt_loss, has_aux=True)(
                train_obj.parameters, teacher_vars["params"], train_obj, inputs
            )
            train_obj.train_state = train_obj.train_state.apply_gradients(grads=grads)
            for mt_losslog in teacher_vars["mt_loss_log"]:
                mt_losslog.update(batch, prediction)

        # update teacher params
        teacher_vars["params"] = jax.tree_map(
            lambda a, b: a * (1 - cls.alpha) + b * cls.alpha,
            teacher_vars["params"],
            train_obj.parameters,
        )

        train_obj.variables[cls.var_key] = teacher_vars

        return prediction, train_obj



import importlib.machinery
import logging
import threading
import traceback
import types
from concurrent import futures
from pathlib import Path

import grpc
import numpy as np
import torch
import typer
from predict import *
from predict import _normalize

from . import proto
from .common import TokenValidationInterceptor, decode_image

app = typer.Typer(pretty_exceptions_enable=False)

_MAX_MSG_SIZE = 1024 * 1024 * 128


def process_input(request: proto.DetectionRequest):
    pixels = request.image_data.pixels
    settings = request.detection_settings

    image = decode_image(pixels).squeeze(0)

    return image


def process_result(masks, image):
    import cv2
    from skimage.measure import regionprops

    response = proto.DetectionResponse()

    for rp in regionprops(masks):
        mask = rp.image.astype("uint8")
        c, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        c = np.array(c[0], dtype=float).squeeze(1)
        c = c + np.array([rp.bbox[1], rp.bbox[0]])
        c = c - 0.5

        scored_roi = proto.ScoredROI(
            score=1.0,
            roi=proto.ROI(
                polygon=proto.Polygon(points=[proto.Point(x=p[0], y=p[1]) for p in c]),
            ),
        )

        response.detections.append(scored_roi)

    return response


class SegformerServicer(proto.LacssServicer):
    def __init__(self, gpu, model_path, model_path2):
        self.model_path = model_path
        self.model_path2 = model_path2
        self.device = "cuda:0" if gpu else "cpu"
        self._lock = threading.RLock()

    def predict(self, img_data):
        hflip_tta = HorizontalFlip()
        vflip_tta = VerticalFlip()

        if img_data.shape[-1] == 1:
            img_data = np.repeat(img_data, 3, axis=-1)
        elif img_data.shape[-1] == 2:
            raise ValueError("Model accept either 1-channel or 3-channel images.")

        H, W = img_data.shape[:2]
        H1 = (H - 1) // 32 * 32 + 32
        W1 = (W - 1) // 32 * 32 + 32
        img_data = np.pad(img_data, [[0, H1 - H], [0, W1 - W], [0, 0]])

        img_data = _normalize(img_data)
        img_data = img_data.astype("float32")
        img_data = np.moveaxis(img_data, -1, 0)
        img_data = img_data[None, ...]

        model = torch.load(self.model_path, map_location=self.device)
        model.eval()

        img_data = torch.from_numpy(img_data).to(self.device)
        img_size = img_data.shape[-1] * img_data.shape[-2]

        if img_size < 1150000 and 900000 < img_size:
            overlap = 0.5
        else:
            overlap = 0.6

        with torch.no_grad():
            img0 = img_data
            outputs0 = sliding_window_inference(
                img0,
                512,
                4,
                model,
                padding_mode="reflect",
                mode="gaussian",
                overlap=overlap,
                device="cpu",
            )
            outputs0 = outputs0.cpu().squeeze()

            if img_size < 2000 * 2000 or img_size > 5000 * 5000:

                model.load_state_dict(
                    torch.load(self.model_path2, map_location=self.device)
                )
                model.eval()

                img2 = hflip_tta.apply_aug_image(img_data, apply=True)
                outputs2 = sliding_window_inference(
                    img2,
                    512,
                    4,
                    model,
                    padding_mode="reflect",
                    mode="gaussian",
                    overlap=overlap,
                    device="cpu",
                )
                outputs2 = hflip_tta.apply_deaug_mask(outputs2, apply=True)
                outputs2 = outputs2.cpu().squeeze()

                outputs = torch.zeros_like(outputs0)
                outputs[0] = (outputs0[0] + outputs2[0]) / 2
                outputs[1] = (outputs0[1] - outputs2[1]) / 2
                outputs[2] = (outputs0[2] + outputs2[2]) / 2

            else:
                # Hflip TTA
                img2 = hflip_tta.apply_aug_image(img_data, apply=True)
                outputs2 = sliding_window_inference(
                    img2,
                    512,
                    4,
                    model,
                    padding_mode="reflect",
                    mode="gaussian",
                    overlap=overlap,
                    device="cpu",
                )
                outputs2 = hflip_tta.apply_deaug_mask(outputs2, apply=True)
                outputs2 = outputs2.cpu().squeeze()
                img2 = img2.cpu()

                ##################
                #                #
                #    ensemble    #
                #                #
                ##################

                model.load_state_dict(
                    torch.load(args.model_path2, map_location=args.device)
                )
                model.eval()

                img1 = img_data
                outputs1 = sliding_window_inference(
                    img1,
                    512,
                    4,
                    model,
                    padding_mode="reflect",
                    mode="gaussian",
                    overlap=overlap,
                    device="cpu",
                )
                outputs1 = outputs1.cpu().squeeze()

                # Vflip TTA
                img3 = vflip_tta.apply_aug_image(img_data, apply=True)
                outputs3 = sliding_window_inference(
                    img3,
                    512,
                    4,
                    model,
                    padding_mode="reflect",
                    mode="gaussian",
                    overlap=overlap,
                    device="cpu",
                )
                outputs3 = vflip_tta.apply_deaug_mask(outputs3, apply=True)
                outputs3 = outputs3.cpu().squeeze()
                img3 = img3.cpu()

                # Merge Results
                outputs = torch.zeros_like(outputs0)
                outputs[0] = (outputs0[0] + outputs1[0] + outputs2[0] - outputs3[0]) / 4
                outputs[1] = (outputs0[1] + outputs1[1] - outputs2[1] + outputs3[1]) / 4
                outputs[2] = (outputs0[2] + outputs1[2] + outputs2[2] + outputs3[2]) / 4

            pred_mask = post_process(outputs.squeeze(0).cpu().numpy(), self.device)

            return pred_mask

    def RunDetection(self, request, context):
        with self._lock:

            logging.info(f"Received message of size {request.ByteSize()}")

            try:
                image = process_input(request)

                if image.ndim == 4:
                    raise ValueError(
                        f"Model does not support 3D input, got shape {image.shape}"
                    )

                logging.info(f"received image {image.shape}")

                preds = self.predict(image)

                response = process_result(preds, image)

                logging.info(f"Reply with message of size {response.ByteSize()}")

                return response

            except ValueError as e:

                logging.error(repr(e))

                logging.error(traceback.format_exc())

                context.abort(grpc.StatusCode.INVALID_ARGUMENT, repr(e))

            except Exception as e:

                logging.error(repr(e))

                logging.error(traceback.format_exc())

                context.abort(
                    grpc.StatusCode.UNKNOWN, f"prediction failed with error: {repr(e)}"
                )

    def RunDetectionStream(self, request_iterator, context):
        with self._lock:
            request = proto.DetectionRequest()

            for next_request in request_iterator:

                if next_request.image_data.HasField("pixels"):
                    request.image_data.pixels.CopyFrom(next_request.image_data.pixels)

                if next_request.image_data.HasField("image_annotation"):
                    request.image_data.image_annotation.CopyFrom(
                        next_request.image_data.image_annotation
                    )

                if next_request.HasField("detection_settings"):
                    request.detection_settings.CopyFrom(next_request.detection_settings)

                if request.image_data.HasField("pixels"):
                    yield self.RunDetection(request, context)


@app.command()
def main(
    port: int = 50051,
    workers: int = 10,
    ip: str = "0.0.0.0",
    local: bool = False,
    token: bool | None = None,
    debug: bool = False,
    compression: bool = True,
    gpu: bool = True,
    model_path: Path = "./main_model.pt",
    model_path2: Path = "./sub_model.pth",
):
    logging.basicConfig(level=logging.DEBUG if debug else logging.INFO)

    print("server starting ...")

    # loader = importlib.machinery.SourceFileLoader('predict', str(predict_py_file))
    # mod = types.ModuleType( loader.name )
    # loader.exec_module(mod)

    if token is None:
        token = not local
    if token:
        import secrets

        token_str = secrets.token_urlsafe(64)

        print()
        print("COPY THE TOKEN BELOW FOR ACCESS.")
        print("=======================================================================")
        print(f"{token_str}")
        print("=======================================================================")
        print()
    else:
        token_str = None

    server = grpc.server(
        futures.ThreadPoolExecutor(max_workers=workers),
        compression=grpc.Compression.Gzip
        if compression
        else grpc.Compression.NoCompression,
        interceptors=(TokenValidationInterceptor(token_str),),
        options=(("grpc.max_receive_message_length", _MAX_MSG_SIZE),),
    )

    proto.add_LacssServicer_to_server(
        SegformerServicer(gpu, model_path, model_path2),
        server,
    )

    if local:
        server.add_secure_port(f"127.0.0.1:{port}", grpc.local_server_credentials())
    else:
        server.add_insecure_port(f"{ip}:{port}")

    logging.info(f"lacss_server: listening on port {port}")

    print("server starting ... ready")

    server.start()
    server.wait_for_termination()


if __name__ == "__main__":
    app()



""" 
Attributes:
    model_urls: URLs for build-in pretrain models. e.g model_urls["default"].
"""
from __future__ import annotations

import logging
import math
import os
import time
from dataclasses import dataclass
from functools import partial, reduce
from typing import Any, Mapping, Sequence, Tuple

import cv2
import flax.linen as nn
import jax
import jax.numpy as jnp
import numpy as np
from vedo import Volume

from ..modules import Lacss
from ..ops import bboxes_of_patches, crop_and_resize_patches
from ..typing import Array, ArrayLike
from ..utils import load_from_pretrained

Shape = Sequence[int]

logger = logging.getLogger(__name__)


def _ensemble_model_fn(module, params, image) -> dict:
    """A JIT compiled function to call Lacss model ensemble
    if params is a sequence (not a dict), assuming a model ensemble and
    generate ensembed prediction.
    """
    from lacss.modules.lpn import generate_predictions

    model_ = module.copy()
    model_.segmentor = None
    model_.segmentor_3d = None

    lpn_out, seg_features = [], []
    for p in params:
        model_out = jax.jit(model_.apply)(dict(params=p), image)

        lpn_out.append(model_out["detector"])

        seg_features.append(model_out["seg_features"])

    lpn_out = jax.tree_util.tree_map(lambda *x: jnp.stack(x), *lpn_out)

    lpn_out = dict(
        logits=lpn_out["logits"].mean(axis=0),
        regressions=lpn_out["regressions"].mean(axis=0),
        ref_locs=lpn_out["ref_locs"][0],
    )

    lpn_out["pred_locs"] = (
        lpn_out["ref_locs"] + lpn_out["regressions"] * module.detector.feature_scale
    )

    predictions = generate_predictions(module.detector, lpn_out)

    assert len(seg_features) == len(params)

    seg_predictions = []
    if image.ndim == 3 and module.segmentor is not None:
        for x, p in zip(seg_features, params):
            seg_predictions.append(
                jax.jit(module.segmentor.apply)(
                    dict(params=p["segmentor"]), x, predictions["locations"]
                )["predictions"]
            )

    elif image.ndim == 4 and module.segmentor_3d is not None:
        seg_predictions = []
        for x, p in zip(seg_features, params):
            seg_predictions.append(
                jax.jit(module.segmentor_3d.apply)(
                    dict(params=p["segmentor_3d"]), x, predictions["locations"]
                )["predictions"]
            )

    if len(seg_predictions) > 0:
        predictions["segmentations"] = jnp.mean(
            jnp.array([x["segmentations"] for x in seg_predictions]),
            axis=0,
        )

        seg_prediction = seg_predictions[0]
        del seg_prediction["segmentations"]

        predictions.update(seg_prediction)

    return predictions


@dataclass
class Predictor:
    """Main class interface for model deployment. This is the only class you
    need if you don't train your own model

    Examples:
        The most common use case is to use a build-in pretrained model.

            import lacss.deploy

            # look up the url of a build-in mode
            url = lacss.deploy.model_urls["default"]

            # create the predictor instance
            predictor = lacss.deploy.Predictor(url)

            # make a prediction
            label = predictor.predict(image)

    Args:
        url: Model file (local or remote) or (module, parameters) tuple. If
            the parameters are a sequence instead of a dictionary, treat it
            as a model ensemble.
        f16: If true, compute in f16 precision (instead of the default f32)
        grid_size: Grid size for large images. Large image are broken down
            into grids to avoid GPU memeory overflow.
        step_size: Should be slightly smaller than grid_size to allow some
            overlap between grids.
        grid_size_3d: Grid_size for 3D input.
        step_size_3d: Step_size for 3D input.
        mask_size: Only use for bbox output. The dimension of the segmentation
            mask output.
        mc_step_size: Only used for 3D contour output. The step size during the
            mesh generation.
        chain_approx: Only used for 2D contour output. Polygon generation rule.
    """

    url: str | os.PathLike | tuple[Lacss, dict]
    f16: bool = False
    grid_size: int = 544
    step_size: int = 480
    cells_per_grid: int = 256
    grid_size_3d: int = 256
    step_size_3d: int = 192
    cells_per_grid_3d: int = 256
    mask_size: int = 36
    mc_step_size: int = 1
    chain_approx: int = cv2.CHAIN_APPROX_SIMPLE

    def __post_init__(self):
        """validate parameters"""

        if isinstance(self.url, tuple):
            module, params = self.url

        else:
            module, params = load_from_pretrained(self.url)

        assert (
            type(module) == Lacss
        ), f"Loaded module is not Lacss, but of {type(module)}"

        # if isinstance(params, Mapping):
        # params = [params]

        # default model config during inference
        module.detector.max_output = self.cells_per_grid

        if module.detector_3d:
            module.detector_3d.max_output = self.cells_per_grid_3d

        if module.segmentor:
            module.segmentor.full_scale_output = True

        if module.segmentor_3d:
            module.segmentor_3d.full_scale_output = False

        self.module = module
        self.params = params

        if self.f16:
            self.params = jax.tree_util.tree_map(
                lambda x: x.astype("float16"),
                self.params,
            )

        assert (
            self.step_size < self.grid_size
        ), f"step_size ({self.step_size}) not smaller than grid_size ({self.grid_size})"
        assert (
            self.grid_size % 32 == 0
        ), f"grid_size ({self.grid_size}) is not divisable by 32"
        assert (
            self.step_size_3d < self.grid_size_3d
        ), f"step_size ({self.step_size_3d}) not smaller than grid_size ({self.grid_size_3d})"
        assert (
            self.grid_size_3d % 32 == 0
        ), f"grid_size ({self.grid_size_3d}) is not divisable by 32"
        assert (
            self.cells_per_grid % 16 == 0
        ), f"cells_per_grid ({self.cells_per_grid}) is not divisable by 16"
        assert (
            self.cells_per_grid_3d % 16 == 0
        ), f"cells_per_grid_3d ({self.cells_per_grid_3d}) is not divisable by 16"

        self._model_fn = jax.jit(
            self.module.apply
        )  # FIXME this locks down model hyperparameter

    def _resize_image(self, image, target_shape):
        from skimage.transform import resize

        # process the data in float32 to avoid overflow
        # only convert to float16 (if needed) at the end

        image = image.astype("float32")

        image = image / (image.std() + 1e-4)
        padding_value = image.mean()
        image -= padding_value

        orig_shape = image.shape[:-1]

        if target_shape is None:
            target_shape = np.array(orig_shape)

        else:
            target_shape = np.broadcast_to(target_shape, [len(orig_shape)])

            scaling = target_shape / orig_shape

            if np.all(np.abs(scaling - 1) < 0.1):  # ignore small scaling
                target_shape = np.array(orig_shape)

            else:
                image = resize(image, target_shape)

        if self.f16:
            image = image.astype("float16")

        logger.debug(f"resized image data from {orig_shape} to {target_shape}")

        return image, target_shape, padding_value

    def _compute_contours_2d(self, predictions, img_sz, threshold):
        """2D contour using cv2"""
        mask = np.array(predictions["segmentation_is_valid"])

        y0s = np.asarray(predictions["segmentation_y0_coord"])
        x0s = np.asarray(predictions["segmentation_x0_coord"])
        segs = np.asarray(predictions["segmentations"] >= threshold).astype("uint8")

        segs = segs.squeeze(1)

        polygons, bboxes = [], []
        for k in range(segs.shape[0]):
            if mask[k]:
                c, _ = cv2.findContours(segs[k], cv2.RETR_EXTERNAL, self.chain_approx)
                max_len_element = reduce(
                    lambda a, b: a if len(a) >= len(b) else b,
                    c,
                    np.zeros([0, 1, 2], dtype=int),
                )

                polygon = max_len_element.squeeze(1).astype(float) + 0.5

                polygon += [x0s[k], y0s[k]]

                if len(polygon) == 0:
                    mask[k] = False
                else:
                    box = np.r_[polygon.min(axis=0), polygon.max(axis=0)]
                    box = box[[1, 0, 3, 2]]
                    if (box[:2] >= img_sz).any() or (box[2:] <= 0).any():
                        mask[k] = False
                    else:
                        polygons.append(polygon)
                        bboxes.append(box)

        assert np.count_nonzero(mask) == len(polygons)

        assert np.count_nonzero(mask) == len(bboxes)

        scores = np.array(predictions["scores"])[mask]

        bboxes = np.array(bboxes).reshape(-1, 4)

        return scores, polygons, bboxes

    def _compute_contours_3d(self, predictions, img_sz, threshold):
        meshes, bboxes = [], []

        mask = np.array(predictions["segmentation_is_valid"])

        z0s = np.asarray(predictions["segmentation_z0_coord"])
        y0s = np.asarray(predictions["segmentation_y0_coord"])
        x0s = np.asarray(predictions["segmentation_x0_coord"])
        segs = np.asarray(predictions["segmentations"] > threshold)

        assert segs.ndim == 4

        for k in range(segs.shape[0]):
            if mask[k]:
                mesh = (
                    Volume(segs[k], spacing=[2, 2, 2])  # model output is 2x binned
                    .isosurface(value=0.5, flying_edges=True)
                    .shift(z0s[k], y0s[k], x0s[k])
                )

                box = np.r_[mesh.bounds()[::2], mesh.bounds()[1::2]]

                if (box[:3] >= img_sz).any() or (box[3:] <= 0).any():
                    mask[k] = False

                else:
                    meshes.append(mesh)
                    bboxes.append(box)

        assert np.count_nonzero(mask) == len(meshes)

        assert np.count_nonzero(mask) == len(bboxes)

        scores = np.array(predictions["scores"])[mask]

        bboxes = np.array(bboxes).reshape(-1, 6)

        return scores, meshes, bboxes

    def _compute_contours(self, predictions, img_sz, threshold):
        if len(img_sz) == 3:
            return self._compute_contours_3d(predictions, img_sz, threshold)
        else:
            return self._compute_contours_2d(predictions, img_sz, threshold)

    def _call_model(
        self, patch, score_threshold, size_threshold, threshold, padding_value
    ):
        """Core inference routine:
        1. Pad input image to regular size and call model function
        2. clean up results
        3. compute contours
        """
        logger.debug(f"process image patch of {patch.shape}")

        # pad image to fixed input sizes
        patch_sz = patch.shape[:-1]
        dim = patch.ndim - 1

        if dim == 2:  # 2D input
            assert (
                max(patch_sz) <= self.grid_size
            ), f"attempted to call model with image shape ({patch_sz}) that exceeded limit."
            h, w = patch_sz
            padding_shape = [
                [0, self.grid_size - h],
                [0, self.grid_size - w],
                [0, 0],
            ]

        else:
            assert (
                max(patch_sz) <= self.grid_size_3d
            ), f"attempted to call model with image shape exceed limit."
            d, h, w = patch_sz
            padding_shape = [
                [0, self.grid_size_3d - d],
                [0, self.grid_size_3d - h],
                [0, self.grid_size_3d - w],
                [0, 0],
            ]

        patch = np.pad(patch, padding_shape, constant_values=padding_value)

        # make 3-ch
        if patch.shape[-1] == 1:
            patch = np.repeat(patch, 3, axis=-1)

        elif patch.shape[-1] == 2:
            patch = np.stack([patch, np.zeros_like(patch[..., :1])], axis=-1)

        logger.debug(f"pad patch to shape: {patch.shape}")

        if isinstance(self.params, Mapping):
            predictions = self._model_fn(dict(params=self.params), patch)["predictions"]
        else:  # handle ensemble
            predictions = _ensemble_model_fn(self.module, self.params, patch)[
                "predictions"
            ]

        # clean up
        mask = predictions["segmentation_is_valid"]

        logger.debug(f"obtained model prediction of {np.count_nonzero(mask)} cells")

        mask &= predictions["scores"] >= score_threshold
        if size_threshold > 0:
            areas = jnp.count_nonzero(
                predictions["segmentations"] > threshold,
                axis=(1, 2, 3),
            )

            mask &= areas >= size_threshold

        predictions["segmentation_is_valid"] = mask

        logger.debug(f"found {np.count_nonzero(mask)} cells after clean up")

        # additional computation
        scores, contours, bboxes = self._compute_contours(
            predictions, patch_sz, threshold
        )

        logger.debug(f"done converting results to contours.")

        assert scores.shape[0] == len(bboxes)
        assert scores.shape[0] == len(contours)

        return scores, contours, bboxes

    def _mark_edge_instances(self, bboxes, pos, patch_sz, img_sz):
        bboxes = np.array(bboxes)
        removal = np.zeros([bboxes.shape[0]], dtype=bool)

        dim = len(img_sz)

        removal |= ((bboxes[:, :dim] <= 1) & (pos > 0)).any(axis=-1)
        removal |= ((bboxes[:, dim:] + 1 >= patch_sz) & (pos + patch_sz < img_sz)).any(
            axis=-1
        )

        return ~removal

    def _process_image(
        self, image, score_threshold, size_threshold, threshold, padding_value
    ):
        """process a 2d/3d image in grids"""
        img_sz = image.shape[:-1]
        img_dim = len(img_sz)

        if img_dim == 2:
            gs, ss = (self.grid_size, self.step_size)
        else:
            gs, ss = (self.grid_size_3d, self.step_size_3d)

        grid_positions = [slice(0, max(d - (gs - ss), 1), ss) for d in img_sz]
        grid_positions = np.moveaxis(np.mgrid[grid_positions], 0, -1)
        grid_positions = grid_positions.reshape(-1, img_dim)

        is_single_grid = grid_positions.shape[0] == 1

        all_scores, all_contours, all_bboxes, all_masks = [], [], [], []
        for pos in grid_positions:
            logger.debug(f"patch position is {pos}")

            slices = (slice(x, x + gs) for x in pos)
            patch = image.__getitem__(tuple(slices))
            patch_sz = patch.shape[:-1]

            scores, contours, bboxes = self._call_model(
                patch,
                score_threshold,
                size_threshold,
                threshold,
                padding_value,
            )

            logger.debug(f"processing patch results")

            # mask = self._mark_edge_instances(bboxes, pos, patch_sz, img_sz)
            mask = np.ones([bboxes.shape[0]], dtype=bool)

            bboxes += np.r_[pos, pos]

            if img_dim == 3:
                contours = [c.shift(pos[0], pos[1], pos[2]) for c in contours]

            else:
                contours = [p + pos[::-1] for p in contours]

            all_scores.append(scores)
            all_contours.append(contours)
            all_bboxes.append(bboxes)
            all_masks.append(mask)

        logger.debug(f"concatenate all results")

        scores = np.concatenate(all_scores)

        if is_single_grid and len(self.params) == 1:
            asort = None
        else:
            asort = np.argsort(scores)[::-1]

        return dict(
            scores=scores,
            mask=np.concatenate(all_masks),
            bboxes=np.concatenate(all_bboxes),
            contours=sum(all_contours, []),
            asort=asort,
        )

    def _non_max_supression(self, predictions, nms_iou):
        from lacss.ops import box_iou_similarity

        if nms_iou <= 0 or nms_iou >= 1:
            return

        selected = predictions["mask"]
        asort = predictions["asort"]
        boxes = predictions["bboxes"]

        if asort is not None:
            boxes = boxes[asort]
            selected = selected[asort]

        sm = np.array(box_iou_similarity(boxes, boxes))
        sm = np.triu(sm, k=1)

        assert len(selected) == boxes.shape[0]

        for k in range(len(selected)):
            if selected[k]:
                selected &= sm[k] < nms_iou

        # inverse selected` to the original order
        if asort is not None:
            selected = selected[asort.argsort()]

        predictions["mask"] = selected

    def _post_process(self, predictions, scaling, nms_iou):
        self._non_max_supression(
            predictions=predictions,
            nms_iou=nms_iou,
        )

        mask = predictions["mask"]

        scores = predictions["scores"][mask]

        bboxes = predictions["bboxes"][mask] / np.r_[scaling, scaling]

        if len(scaling) == 2:
            contours = [
                p / scaling[::-1] for p, bm in zip(predictions["contours"], mask) if bm
            ]

        else:
            contours = []
            for mesh, bm in zip(predictions["contours"], mask):
                if bm:
                    mesh = mesh.scale(1 / scaling, origin=False)
                    divisions = np.ceil(
                        (mesh.bounds()[1::2] - mesh.bounds()[::2]) / self.mc_step_size
                    ).astype(int)
                    mesh = mesh.decimate_binned(divisions)

                    contours.append(mesh)

        return dict(
            scores=scores,
            bboxes=bboxes,
            contours=contours,
        )

    def _to_contour_format(self, preds, img_sz):
        if len(img_sz) == 3:
            simple_contours = []
            for mesh in preds["contours"]:
                simple_contours.append(
                    dict(
                        verts=mesh.vertices,
                        faces=np.array(mesh.cells),
                    )
                )
            preds["contours"] = simple_contours

        return dict(
            pred_scores=preds["scores"],
            pred_bboxes=preds["bboxes"],
            pred_contours=preds["contours"],
        )

    def _to_label_format(self, preds, img_sz):
        label = np.zeros(img_sz, dtype="uint16")

        if len(img_sz) == 2:
            polygons = preds["contours"]

            color = len(polygons)
            for polygon in polygons[::-1]:
                cv2.fillPoly(label, [polygon.astype(int)], color)  # type: ignore
                color -= 1

        else:
            meshes = preds["contours"]

            d, h, w = img_sz

            color = 1
            for mesh in meshes[::-1]:
                origin = np.floor(mesh.bounds()[::2]).astype(int)
                origin = np.clip(origin, 0, (d - 1, h - 1, w - 1))
                max_size = np.array(label.shape) - origin

                vol = mesh.binarize(
                    values=(color, 0),
                    spacing=[1, 1, 1],
                    origin=origin,
                )

                vol_d = vol.tonumpy()[: max_size[0], : max_size[1], : max_size[2]]
                size = tuple(vol_d.shape)

                region = label[
                    origin[0] : origin[0] + size[0],
                    origin[1] : origin[1] + size[1],
                    origin[2] : origin[2] + size[2],
                ]
                region[...] = np.maximum(region, vol_d)

                color = color + 1

        return dict(
            pred_scores=preds["scores"],
            pred_label=label,
        )

    def _to_bbox_format(self, preds, img_sz):
        n_det = len(preds["contours"])

        if len(img_sz) == 2:
            masks = np.zeros([n_det, self.mask_size, self.mask_size], dtype="uint8")

            for k in range(n_det):
                p = preds["contours"][k]
                box = preds["bboxes"][k]
                p = (p - box[:2]) / (box[2:] - box[:2]) * self.mask_size
                cv2.fillPoly(masks[k], [p.astype(int)], 255)  # type: ignore

            return dict(
                pred_scores=preds["scores"],
                pred_bboxes=preds["bbxoxes"],
                pred_masks=masks,
            )

        else:
            masks = []

            for k in range(n_det):
                mesh = preds["contours"][k]
                box = preds["bboxes"][k]

                masks.append(
                    mesh.binarize(
                        values=(255, 0),
                        dims=[self.mask_size] * 3,
                        origin=box[:3],
                    )
                    .tonumpy()
                    .astype("uint8")
                )

            return dict(
                pred_scores=preds["scores"],
                pred_bboxes=preds["bboxes"],
                pred_masks=np.stack(masks),
            )

    def predict(
        self,
        image: ArrayLike,
        *,
        output_type: str = "label",
        reshape_to: int | tuple[int] | np.ndarray | None = None,
        min_area: float = 0,
        score_threshold: float = 0.5,
        segmentation_threshold: float = 0.5,
        nms_iou: float = 0.4,
        remove_out_of_bound: bool | None = None,
    ) -> dict:
        """Predict segmentation.

        Args:
            image: A ndarray of (h,w,c) or (d,h,w,c) format. c must be 1-3

        Keyword Args:
            output_type: "label" | "contour" | "bbox"
            reshape_to: If not None, the input image will be resized internally before send
                to the model. The results will be resized back to the scale of the orginal input image.
            min_area: Minimum area of a valid prediction.
            score_threshold: Min score needed to be included in the output.
            segmentation_threshold: Threshold value for segmentation
            nms_iou: IOU threshold value for non-max-supression post-processing

        Returns:
            For "label" output:

                - pred_scores: The prediction scores of each instance.
                - pred_label: a 2D image label. 0 is background.

            For "contour" output:

                - pred_scores: The prediction scores of each instance.
                - pred_contours: a list of polygon arrays in x-y format.

            For "bbox" output (ie MaskRCNN):

                - pred_scores: The prediction scores of each instance.
                - pred_bboxes: The bounding-boxes of detected instances in y0x0y1x1 format
                - pred_masks:  A 3d array representing (rescaled) segmentation mask within bboxes
        """
        image = np.asarray(image)

        logger.debug(f"started prediction with image {image.shape}")
        start_time = time.time()

        if remove_out_of_bound is not None:
            import warnings

            warnings.warn("remove_out_of_bound is deprecated", DeprecationWarning, 2)

        if image.ndim == 2 or image.shape[-1] > 3:
            logger.warning("input seems to have no channel dim. Add one")
            image = image[..., None]

        assert image.ndim == 3 or image.ndim == 4, f"illegal image dim {image.shape}"

        assert image.shape[-1] <= 3, f"input image has more than 3 channels"

        if not output_type in ("bbox", "label", "contour", "_raw"):
            raise ValueError(
                f"output_type should be 'bbox'|'label'|'contour'. Got {output_type} instead."
            )

        img_shape = image.shape[:-1]
        dim = len(img_shape)

        image, reshape_to, padding_value = self._resize_image(image, reshape_to)

        scaling = reshape_to / img_shape

        logger.debug(f"done preprocessing")

        predictions = self._process_image(
            image,
            score_threshold,
            min_area * np.prod(scaling),
            math.log(segmentation_threshold / (1 - segmentation_threshold)),
            padding_value,
        )

        logger.debug(f"post processing ...")
        predictions = self._post_process(predictions, scaling, nms_iou)

        # generate outputs
        if output_type == "_raw":
            outputs = predictions

        elif output_type == "contour":
            outputs = self._to_contour_format(predictions, img_shape)

        elif output_type == "bbox":
            outputs = self._to_bbox_format(predictions, img_shape)

        else:
            outputs = self._to_label_format(predictions, img_shape)

        elapsed = (time.time() - start_time) * 1000
        logger.debug(f"done prediction in {elapsed:.2f} ms")

        return outputs



"""Vision Transformer.
Based on scenic implementation: https://github.com/google-research/scenic/
"""

from typing import Any, Callable, Optional, Sequence

import flax.linen as nn
import jax
import jax.numpy as jnp
import numpy as np

from ..typing import Array, ArrayLike
from .common import (
    FFN,
    DropPath,
    IdentityLayer,
    PositionEmbedding1D,
    PositionEmbedding2D,
)

Initializer = Callable[[jnp.ndarray, Sequence[int], jnp.dtype], jnp.ndarray]


class MAPHead(nn.Module):
    """Multihead Attention Pooling."""

    mlp_dim: Optional[int] = None  # Defaults to 4x input dim
    num_heads: int = 12

    @nn.compact
    def __call__(self, x):
        x_shape = x.shape
        x = x.reshape(-1, x_shape[-3] * x_shape[-2], x_shape[-1])
        probe = self.param(
            "probe",
            nn.initializers.xavier_uniform(),
            (1, 1, x.shape[-1]),
            self.param_dtype,
        )
        probe = probe.astype(x.dtype)
        probe = jnp.tile(probe, [x.shape[0], 1, 1])

        x = nn.MultiHeadDotProductAttention(
            num_heads=self.num_heads,
            kernel_init=nn.initializers.xavier_uniform(),
        )(probe, x)

        y = nn.LayerNorm()(x)
        x = x + FFN(
            self.mlp_dim,
            dropout_rate=0.0,
        )(y, deterministic=True)

        return x[:, 0].reshape(*x_shape[:-3], -1)


class VitEncoder(nn.Module):
    """Transformer Encoder.

    Attributes:
      n_layers: Number of layers.
      n_heads: The number of heads for multi-head self-attention.
      dim_ffn: Dimension of the mlp on top of attention block.
      pos_embedding_type: The type of positional embeddings to add to the
        input tokens.
      dropout: Dropout rate.
      stochastic_depth: probability of dropping a layer linearly grows
        from 0 to the provided value. Our implementation of stochastic depth
        follows timm library, which does per-example layer dropping and uses
        independent dropping patterns for each skip-connection.
    """

    n_layers: int = 6
    n_heads: int = 12
    dim_ffn: int | None = None
    pos_embedding_type: str = "learned_1d"
    dropout: float = 0.1
    attention_dropout: float = 0.1
    stochastic_depth: float = 0.0

    @nn.compact
    def __call__(
        self,
        inputs: ArrayLike,
        cls_token: ArrayLike | None = None,
        *,
        training: bool = False,
    ):
        """Applies Transformer model on the inputs.

        Args:
          inputs: Input tokens of shape [..., num_tokens, channels].
          cls_token: addtional class token if not None
          training: If in training mode, dropout and stochastic depth is applied.

        Returns:
          Encoded tokens and cls_token
        """

        input_shape = inputs.shape
        H, W, C = input_shape[-3:]

        x = inputs.reshape((-1,) + (H * W, C))

        if self.pos_embedding_type == "learned_1d":
            x = x + PositionEmbedding1D()(x)

        elif self.pos_embedding_type == "sinusoidal_2d":
            pos = PositionEmbedding2D(
                posemb_init=None,
            )(input_shape)
            x = x + pos.reshape(-1, C)

        elif self.pos_embedding_type == "learned_2d":
            pos = PositionEmbedding2D()(input_shape)
            x = x + pos.reshape(-1, C)

        elif self.pos_embedding_type != "none":
            raise ValueError("invalid pos_embedding_type {self.pos_embedding_type}")

        if cls_token:
            assert cls_token.shape == input_shape[:-3] + (
                C,
            ), f"invalid class token shape {cls_token.shape}"
            x = jnp.c_[x, cls_token.reshape(-1, 1, C)]

        # x = nn.Dropout(rate=self.dropout)(x, deterministic=not training)

        for lyr in range(self.n_layers):
            shortcut = x
            x = nn.Layernorm()(x)
            x = nn.MultiHeadAttention(
                num_heads=self.n_heads,
                dropout_rate=self.attention_dropout,
                kernel_init=nn.initializers.xavier_uniform(),
                broadcast_dropout=False,
                name=f"attention_{lyr}",
            )(x, deterministic=not training)

            x = FFN(dim=self.dim_ffn, dropout_rate=self.dropout, name="ffn_{lyr}")(
                x, deterministic=not training
            )

            droppath_rate = lyr / self.n_layers * self.stochastic_depth
            x = DropPath(droppath_rate)(x, deterministic=not training)

            x = shortcut + x

        if cls_token:
            cls_token = x[..., -1, :].reshape(input_shape[:-3] + (C,))
            x = x[..., :-1, :].reshape(input_shape)
            return x, cls_token
        else:
            x = x.reshape(input_shape)
            return x, None


class ViT(nn.Module):
    """Vision Transformer model.

    Attributes:
      dim: Size of the hidden state of the output of model's stem.
      dim_ffn: Dimension of the mlp on top of attention block.
      n_layers: Number of layers.
      n_heads: Number of self-attention heads.
      patch_size: specify the patch dimension for tokenization
      pos_embedding_type: The type of positional embeddings to add to the
          tokens at the beginning of the transformer encoder.
      dropout: Dropout rate.
      attention_dropout: Dropout for attention heads.

      n_cls: Number of output classes.
      representation_size: Size of the representation layer in the model's head.
          if None, we skip the extra projection + tanh activation at the end.
      classifier: type of the classifier layer. Options are 'gap', 'gmp', 'gsp',
          'token', or None to return encoder features
    """

    # transformer parameters
    dim: int = 256
    dim_ffn: int | None = None
    n_layers: int = 8
    n_heads: int = 12
    patch_size: tuple[int, int] = (8, 8)
    pos_embedding_type: str = "learned_1d"
    dropout: float = 0.1
    attention_dropout: float = 0.1
    stochastic_depth: float = 0.0

    # classifier parameters
    n_cls: int = -1
    representation_size: Optional[int] = None
    classifier_type: str | None = None

    @nn.compact
    def __call__(self, x: jnp.ndarray, *, training: bool = False):
        """ViT model

        Args:
          inputs: Input [..., H, W, C].
          training: If in training mode, dropout and stochastic depth is applied.

        Returns:
          features: [..., H//patch_size, W//patch_size, n_dim_hidden]
        """
        input_shape = x.shape

        x = x.reshape((-1,) + input_shape[-3:])  # [b, h, w, c]

        fh, fw = self.patch_size
        x = nn.Conv(
            self.dim,
            (fh, fw),
            strides=(fh, fw),
            padding="VALID",
            name="embedding",
        )(x)

        output_shape = x.shape

        # x = jnp.reshape(x.shape[0], -1, x.shape[-1]) # merge H and W

        # Add a class token if suitable
        if self.classifier_type == "token":
            cls_token = self.param(
                "cls", nn.initializers.zeros, (1, x.shape[-1]), self.param_dtype
            )
            cls_toekn = cls_token.astype(x.dtype)
            cls_token = jnp.tile(cls_token, [x.shape[0], 1])
        else:
            cls_token = None

        x, cls_token = VitEncoder(
            dim_ffn=self.dim_ffn,
            n_layers=self.n_layers,
            n_heads=self.n_heads,
            pos_embedding_type=self.pos_embedding_type,
            dropout=self.dropout,
            attention_dropout=self.attention_dropout,
            stochastic_depth=self.stochastic_depth,
            name="Transformer",
        )(x, cls_token, training=training)

        if self.classifier_type is None:
            return x.reshape(output_shape)

        if self.classifier_type in ("token", "0"):
            x = cls_token

        elif self.classifier_type in ("gap", "gmp", "gsp"):
            fn = {"gap": jnp.mean, "gmp": jnp.max, "gsp": jnp.sum}[self.classifier_type]
            x = fn(x, axis=1)

        elif self.classifier_type == "map":
            x = MAPHead(
                num_heads=self.n_heads,
                mlp_dim=self.dim_ffn,
            )(x)

        else:
            raise ValueError(f"Unknown classifier {self.classifier_type}")

        if self.representation_size is not None:
            x = nn.Dense(self.representation_size, name="pre_logits")(x)
            x = nn.tanh(x)
        else:
            x = IdentityLayer(name="pre_logits")(x)

        # If self.n_cls <= 0, we just return the representations
        if self.n_cls > 0:
            x = nn.Dense(
                self.n_cls,
                kernel_init=nn.initializers.zeros,
                name="output_projection",
            )(x)

        output = x.reshape(input_shape[:-3] + x.shape[-1:])

        return output



from __future__ import annotations

import flax.linen as nn
import jax

from ..typing import Any, Array, ArrayLike, Sequence
from .common import FFN, PositionEmbedding1D

jnp = jax.numpy


class VideoIntegrator(nn.Module):
    n_heads: int = 4
    dropout: float = 0.2
    ff_dropout: float = 0.2
    dtype: Any = None

    @nn.compact
    def __call__(
        self,
        feature: Sequence[ArrayLike],
        video_refs: tuple[Sequence[ArrayLike], ArrayLike | None],
        deterministic: bool = True,
    ) -> Sequence[Array]:
        ref_feature, ref_mask = video_refs

        out = []
        for k in range(4):
            x = jnp.asarray(feature[k])[..., None, :]

            y = jnp.moveaxis(ref_feature[k], 0, -2)
            pos_embedding = PositionEmbedding1D()(y)

            x = nn.LayerNorm(epsilon=1e-6, dtype=self.dtype)(x)
            x = nn.MultiHeadDotProductAttention(
                self.n_heads,
                dropout_rate=self.dropout,
                dtype=self.dtype,
            )(
                inputs_q=x,
                inputs_k=y + pos_embedding,
                inputs_v=y,
                deterministic=deterministic,
                sow_weights=True,
                mask=ref_mask,
            )

            x = FFN(dropout_rate=self.ff_dropout, dtype=self.dtype)(
                x, deterministic=deterministic
            )

            x = x.squeeze(-2)

            out.append(feature[k] + x)

        return out



import threading

import biopb.image as proto
import grpc
import numpy as np
from biopb.image.utils import deserialize_to_numpy

_AUTH_HEADER_KEY = "authorization"


def decode_image(pixels: proto.Pixels) -> np.ndarray:
    if pixels.size_t > 1:
        raise ValueError("Image data has a non-singleton T dimension.")

    if pixels.size_c > 3:
        raise ValueError("Image data has more than 3 channels.")

    np_img = deserialize_to_numpy(pixels)

    return np_img


class TokenValidationInterceptor(grpc.ServerInterceptor):
    def __init__(self, token):
        def abort(ignored_request, context):
            context.abort(grpc.StatusCode.UNAUTHENTICATED, "Invalid token signature")

        self._abort_handler = grpc.unary_unary_rpc_method_handler(abort)
        self.token = token

    def intercept_service(self, continuation, handler_call_details):
        expected_metadata = (_AUTH_HEADER_KEY, f"Bearer {self.token}")
        if (
            self.token is None
            or expected_metadata in handler_call_details.invocation_metadata
        ):
            return continuation(handler_call_details)
        else:
            return self._abort_handler


class LacssServicerBase(proto.ObjectDetectionServicer, proto.ProcessImageServicer):
    def __init__(self):
        self._lock = threading.RLock()

    def RunDetectionStream(self, request_iterator, context):
        with self._lock:
            request = proto.DetectionRequest()

            for next_request in request_iterator:

                if next_request.image_data.HasField("pixels"):
                    request.image_data.pixels.CopyFrom(next_request.image_data.pixels)

                if next_request.image_data.HasField("image_annotation"):
                    request.image_data.image_annotation.CopyFrom(
                        next_request.image_data.image_annotation
                    )

                if next_request.HasField("detection_settings"):
                    request.detection_settings.CopyFrom(next_request.detection_settings)

                if request.image_data.HasField("pixels"):
                    yield self.RunDetection(request, context)



from __future__ import annotations

from functools import partial
from typing import Callable, Sequence, Tuple

import flax.linen as nn
import jax
import jax.numpy as jnp

from ..typing import Any, Array, ArrayLike
from .common import DefaultUnpicklerMixin
from .lpn import generate_predictions


class LPN3D(nn.Module, DefaultUnpicklerMixin):
    """Location 3d detection head

    Attributes:
        n_layers: num of conv layers for feature mixing
        nms_threshold: non-max-supression threshold, if performing nms on detected locations.
        pre_nms_topk: max number of detections to be processed regardless of nms, ignored if negative
        max_output: number of detection outputs
    """

    # network hyperparams
    n_layers: int = 2
    dim: int = 384
    feature_scale: int = 4
    dtype: Any = None

    # detection hyperparams
    nms_threshold: float = 8.0
    pre_nms_topk: int = -1
    max_output: int = 256
    min_score: float = 0.2

    def _block(self, feature: Array) -> dict:
        x = feature

        depth, height, width = x.shape[:-1]

        logits = nn.Conv(1, (1, 1), dtype=self.dtype)(x)
        regressions = nn.Conv(3, (1, 1), dtype=self.dtype)(x)

        ref_locs = (
            jnp.moveaxis(jnp.mgrid[:depth, :height, :width] + 0.5, 0, -1)
            * self.feature_scale
        )
        locs = ref_locs + regressions * self.feature_scale

        return dict(
            regressions=regressions.reshape(-1, 3),
            logits=logits.reshape(-1),
            ref_locs=ref_locs.reshape(-1, 3),
            pred_locs=locs.reshape(-1, 3),
        )

    def _mix_feaures(self, feature: Array) -> Array:
        x = feature
        dim = self.dim
        for _ in range(self.n_layers):
            x = nn.Conv(dim, (3, 3, 3), dtype=self.dtype)(x)
            x = nn.gelu(x)

        return x

    @nn.compact
    def __call__(self, feature: ArrayLike, mask: ArrayLike | None = None) -> dict:
        feature = jnp.asarray(feature)

        x = self._mix_feaures(feature)
        x = self._block(x)
        predictions = generate_predictions(self, x)

        if mask is not None:
            pred_locs = jnp.floor(predictions["locations"]).astype(int)
            pred_locs_ = jnp.clip(pred_locs, 0, jnp.array(mask.shape) - 1)
            masked = jnp.where(
                (pred_locs == pred_locs_).all(axis=1),
                mask[tuple(pred_locs_.transpose())],
                False,
            )
            predictions["scores"] = jnp.where(masked, predictions["scores"], 0)
            # predictions['locations'] = jnp.where(masked[:, None], predictions['locations'], -1)

        return dict(
            detector=x,
            predictions=predictions,
        )



import logging
import traceback
from concurrent import futures
from contextlib import contextmanager
from pathlib import Path
from typing import Iterable

import biopb.image as proto
import grpc
import jax
import numpy as np
import typer
from biopb.image.utils import serialize_from_numpy

from .common import LacssServicerBase, TokenValidationInterceptor, decode_image

logging.basicConfig(
    format="%(asctime)s - %(name)s - %(levelname)s - %(filename)s:%(lineno)d - %(funcName)s() - %(message)s"
)

logger = logging.getLogger(__name__)

app = typer.Typer(pretty_exceptions_enable=False)

_MAX_MSG_SIZE = 1024 * 1024 * 128
_TARGET_CELL_SIZE = 32
_MAX_STREAM_MSG_SIZE = _MAX_MSG_SIZE * 16
MAX_IMG_SIZE = 1024 * 1024 * 1024


def _process_input(request: proto.DetectionRequest, image=None):
    pixels = request.image_data.pixels
    settings = request.detection_settings

    if image is None:
        image = decode_image(pixels)

    physical_size = np.array(
        [
            pixels.physical_size_z
            or pixels.physical_size_x,  # one might set xy but not z
            pixels.physical_size_y,
            pixels.physical_size_x,
        ],
        dtype="float",
    )
    if (physical_size == 0).any():
        physical_size[:] = 1.0

    if settings.HasField("cell_diameter_hint"):
        scaling = _TARGET_CELL_SIZE / settings.cell_diameter_hint * physical_size

    else:
        if physical_size[1] != physical_size[2]:
            raise ValueError("Scaling hint provided, but pixel is not isometric")

        scaling = np.array([settings.scaling_hint or 1.0] * 3, dtype="float")
        scaling[0] *= physical_size[0] / physical_size[1]

    logger.info(f"Requested rescaling factor is {scaling}")

    shape_hint = tuple(np.round(scaling * image.shape[:3]).astype(int))

    if image.shape[0] == 1:  # 2D
        image = image.squeeze(0)
        shape_hint = shape_hint[1:]

    kwargs = dict(
        reshape_to=shape_hint,
        score_threshold=settings.min_score or 0.4,
        min_area=settings.min_cell_area,
        nms_iou=settings.nms_iou or 0.4,
        segmentation_threshold=settings.segmentation_threshold or 0.5,
    )

    return image, kwargs


def _process_result(preds, image) -> proto.DetectionResponse:
    response = proto.DetectionResponse()

    if image.ndim == 3:  # returns polygon

        for contour, score in zip(preds["pred_contours"], preds["pred_scores"]):
            if len(contour) == 0:
                continue

            scored_roi = proto.ScoredROI(
                score=score,
                roi=proto.ROI(
                    polygon=proto.Polygon(
                        points=[proto.Point(x=p[0], y=p[1]) for p in contour]
                    ),
                ),
            )

            response.detections.append(scored_roi)

    else:  # 3d returns Mesh
        for mesh, score in zip(preds["pred_contours"], preds["pred_scores"]):
            scored_roi = proto.ScoredROI(
                score=score,
                roi=proto.ROI(
                    mesh=proto.Mesh(
                        verts=[
                            proto.Point(z=v[0], y=v[1], x=v[2]) for v in mesh["verts"]
                        ],
                        faces=[
                            proto.Mesh.Face(p1=p[0], p2=p[1], p3=p[2])
                            for p in mesh["faces"]
                        ],
                    ),
                ),
            )

            response.detections.append(scored_roi)

    return response


def _process_grid_input(request_iterator: Iterable[proto.DetectionRequest]):
    d, h, w, c = 0, 0, 0, 0
    images, grids = [], []
    request = None
    total_msg_size = 0
    for request in request_iterator:
        total_msg_size += request.ByteSize()
        if total_msg_size > _MAX_STREAM_MSG_SIZE:
            raise ValueError("input message size  {total_msg_size} exceeded limit.")

        pixels = request.image_data.pixels

        image = decode_image(pixels)

        grids.append(
            [
                pixels.offset_z,
                pixels.offset_y,
                pixels.offset_x,
            ]
        )
        images.append(image)

        d = max(d, pixels.offset_z + image.shape[0])
        h = max(d, pixels.offset_z + image.shape[1])
        w = max(d, pixels.offset_z + image.shape[2])
        c = max(c, image.shape[3])

        if d * h * w > MAX_IMG_SIZE:
            raise ValueError(f"input image is too large {(d, h, w)}")

    # empty input iterator
    if request is None:
        return None, None

    assert c <= 3

    full_image = np.zeros([d, h, w, c], dtype=images[0].dtype)
    for image, grid in zip(images, grids):
        full_image[
            grid[0] : grid[0] + image.shape[0],
            grid[1] : grid[1] + image.shape[1],
            grid[2] : grid[2] + image.shape[2],
            : image.shape[3],
        ] = image

    return _process_input(request, full_image)


class LacssServicer(LacssServicerBase):
    def __init__(self, model):
        super().__init__()

        self.model = model

    @contextmanager
    def _lacss_context(self, context):
        try:
            with self._lock:
                yield

        except ValueError as e:
            logger.error(repr(e))

            logger.error(traceback.format_exc())

            context.abort(grpc.StatusCode.INVALID_ARGUMENT, repr(e))

        except Exception as e:

            logger.error(repr(e))

            logger.error(traceback.format_exc())

            context.abort(
                grpc.StatusCode.UNKNOWN, f"prediction failed with error: {repr(e)}"
            )

    def RunDetection(self, request, context):
        logger.info(f"Received message of size {request.ByteSize()}")

        with self._lacss_context(context):
            image, kwargs = _process_input(request)

            logger.info(f"received image {image.shape}")

            preds = self.model.predict(
                image,
                output_type="contour",
                **kwargs,
            )

            response = _process_result(preds, image)

            logger.info(f"Reply with message of size {response.ByteSize()}")

            return response

    def RunDetectionOnGrid(self, request_iterator, context):
        with self._lacss_context(context):
            image, kwargs = _process_grid_input(request_iterator)

            if image is None:
                return proto.DetectionResponse()

            logger.info(f"Received full image of size {image.shape[:-1]}")

            preds = self.model.predict(
                image,
                output_type="contour",
                **kwargs,
            )

            response = _process_result(preds, image)

            logger.info(f"Reply with message of size {response.ByteSize()}")

            return response

    def Run(self, request, context):
        logger.info(f"Received message of size {request.ByteSize()}")

        with self._lacss_context(context):
            image = decode_image(request.image_data.pixels)

            if image.shape[0] == 1:  # 2D
                image = image.squeeze(0)

            logger.info(f"received image {image.shape}")

            label = self.model.predict(image)["pred_label"]

            logger.info(f"Detected {label.max()} cells")

            response = proto.ProcessResponse(
                image_data=proto.ImageData(pixels=serialize_from_numpy(label)),
            )

            logger.info(f"Reply with message of size {response.ByteSize()}")

            return response


def get_predictor(modelpath, f16):
    from .predict import Predictor

    model = Predictor(modelpath, f16=f16)

    logger.info(f"lacss_server: loaded model from {modelpath}")

    model.module.detector.min_score = 0.2
    if model.module.detector_3d:
        model.module.detector_3d.min_score = 0.2

    logger.debug(f"lacss_server: precompile the model")

    _ = model.predict(np.ones([255, 255, 255, 3]), output_type="_raw")
    _ = model.predict(np.ones([544, 544, 3]), output_type="_raw")

    return model


def show_urls():
    from . import model_urls

    print("Pretrained model files:")
    print("==============================")
    for k, v in model_urls.items():
        print(f"{k}: {v}")
    print()


@app.command()
def main(
    modelpath: Path | None = None,
    port: int = 50051,
    workers: int = 10,
    ip: str = "0.0.0.0",
    local: bool = False,
    token: bool | None = None,
    debug: bool = False,
    compression: bool = True,
    f16: bool = False,
):
    logger.setLevel(logging.DEBUG if debug else logging.INFO)

    if modelpath is None:
        show_urls()
        return

    print("server starting ...")

    model = get_predictor(modelpath, f16)

    logger.info(f"lacss_server: default backend is {jax.default_backend()}")

    if jax.default_backend() == "cpu":
        logger.warning(
            f"lacss_server: WARNING: No GPU configuration. This might be very slow ..."
        )

    if token is None:
        token = not local
    if token:
        import secrets

        token_str = secrets.token_urlsafe(64)

        print()
        print("COPY THE TOKEN BELOW FOR ACCESS.")
        print("=======================================================================")
        print(f"{token_str}")
        print("=======================================================================")
        print()
    else:
        token_str = None

    server = grpc.server(
        futures.ThreadPoolExecutor(max_workers=workers),
        compression=grpc.Compression.Gzip
        if compression
        else grpc.Compression.NoCompression,
        interceptors=(TokenValidationInterceptor(token_str),),
        options=(("grpc.max_receive_message_length", _MAX_MSG_SIZE),),
    )

    servicer = LacssServicer(model)
    proto.add_ObjectDetectionServicer_to_server(servicer, server)
    proto.add_ProcessImageServicer_to_server(servicer, server)

    if local:
        server.add_secure_port(f"127.0.0.1:{port}", grpc.local_server_credentials())
    else:
        server.add_insecure_port(f"{ip}:{port}")

    logger.info(f"lacss_server: listening on port {port}")

    print("server starting ... ready")

    server.start()
    server.wait_for_termination()


if __name__ == "__main__":
    app()
